""" Robotica Schedule. """
import asyncio
import logging
from typing import Dict, Any, Set, List  # NOQA

import yaml

from robotica.outputs import Output
from robotica.types import Action

logger = logging.getLogger(__name__)


class Executor:
    def __init__(
            self, loop: asyncio.AbstractEventLoop, config: Dict) -> None:
        self._loop = loop
        self._config = config
        self._outputs = []  # type: List[Output]
        self._lock = asyncio.Lock()

    def add_output(self, output: Output) -> None:
        self._outputs.append(output)

    def is_action_required_for_locations(
            self, locations: Set[str], action: Action) -> bool:

        action_required = [
            output.is_action_required_for_locations(locations, action)
            for output in self._outputs
        ]

        return any(action_required)

    async def do_action(self, locations: Set[str], action: Action) -> None:
        if self.is_action_required_for_locations(locations, action):
            with await self._lock:
                coros = [
                    output.execute(locations, action)
                    for output in self._outputs
                ]
                await asyncio.gather(
                    *coros,
                    loop=self._loop
                )

    async def do_actions(self, locations: Set[str], actions: List[Action]) -> None:
        for action in actions:
            await self.do_action(locations, action)



""" Give verbal message. """
import asyncio
import logging
import shlex
from typing import Dict, List, Set

import yaml

from robotica.outputs import Output
from robotica.types import Action

logger = logging.getLogger(__name__)


class AudioOutput(Output):

    def __init__(self, loop: asyncio.AbstractEventLoop, config: Dict) -> None:
        self._loop = loop
        self._config = config
        self._disabled = self._config['disabled']
        self._locations = self._config.get('locations', {}) or {}

    def start(self) -> None:
        pass

    def stop(self) -> None:
        pass

    def is_action_required_for_locations(self, locations: Set[str], action: Action) -> bool:
        if self._disabled:
            return False

        good_locations = [
            location
            for location in locations
            if location in self._locations
        ]

        if len(good_locations) == 0:
            return False

        if 'message' in action:
            return True

        if 'music' in action:
            return True

        return False

    async def execute(self, locations: Set[str], action: Action) -> None:
        if 'message' in action:
            message = action['message']

            await self.say(
                locations=locations,
                text=message['text'])

        if 'music' in action:
            music = action['music']

            await self.music_play(
                locations=locations,
                play_list=music['play_list'])

    @staticmethod
    async def _execute(cmd_list: List[str], params: Dict[str, str]) -> None:
        for cmd in cmd_list:
            split = [
                value.format(**params) for value in shlex.split(cmd)
            ]
            logger.info("About to execute %s", split)
            process = await asyncio.create_subprocess_exec(*split)
            result = await process.wait()
            if result != 0:
                logger.info("Command %s returned %d", split, result)

    async def _say(self, location: str, text: str) -> None:
        logger.debug("%s: About to consider say '%s'.", location, text)
        location_config = self._locations.get(location, {})
        say_cmd = location_config.get('say_cmd', [])
        if len(say_cmd) == 0:
            return
        logger.debug("%s: About to say '%s'.", location, text)
        await self._music_stop(location)
        await self._play(location, 'prefix')
        await self._execute(say_cmd, {'text': text})
        await self._play(location, 'repeat')
        await self._execute(say_cmd, {'text': text})
        await self._play(location, 'postfix')

    async def say(self, locations: Set[str], text: str) -> None:
        if self._disabled:
            return
        coros = [
            self._say(location, text)
            for location in locations
        ]
        await asyncio.gather(*coros, loop=self._loop)

    async def _play(self, location: str, sound: str) -> None:
        sound_file = self._config['sounds'].get(sound)
        if not sound_file:
            return
        location_config = self._locations.get(location, {})
        play_cmd = location_config.get('play_cmd', [])
        if len(play_cmd) == 0:
            return
        logger.debug("%s: About to play sound '%s'.", location, sound_file)
        await self._execute(play_cmd, {'file': sound_file})

    async def play(self, locations: Set[str], sound: str) -> None:
        if self._disabled:
            return
        coros = [
            self._play(location, sound)
            for location in locations
        ]
        await asyncio.gather(*coros, loop=self._loop)

    async def _music_play(self, location: str, play_list: str) -> None:
        location_config = self._locations.get(location, {})
        music_play_cmd = location_config.get('music_play_cmd', [])
        if len(music_play_cmd) == 0:
            return
        logger.debug("%s: About to play music '%s'.", location, play_list)
        await self._execute(music_play_cmd, {'play_list': play_list})

    async def music_play(self, locations: Set[str], play_list: str) -> None:
        if self._disabled:
            return
        coros = [
            self._music_play(location, play_list)
            for location in locations
        ]
        await asyncio.gather(*coros, loop=self._loop)

    async def _music_stop(self, location: str) -> None:
        location_config = self._locations.get(location, {})
        music_stop_cmd = location_config.get('music_stop_cmd', [])
        if len(music_stop_cmd) == 0:
            return
        logger.debug("%s: About to stop music.", location)
        await self._execute(music_stop_cmd, {})

    async def music_stop(self, locations: Set[str]) -> None:
        if self._disabled:
            return
        coros = [
            self._music_stop(location)
            for location in locations
        ]
        await asyncio.gather(*coros, loop=self._loop)



from typing import Set

from robotica.types import Action


class Output:
    def is_action_required_for_locations(self, locations: Set[str], action: Action) -> bool:
        raise NotImplemented()

    async def execute(self, locations: Set[str], action: Action) -> None:
        raise NotImplemented()



# -*- coding: utf-8 -*-

"""Console script for Robotica."""
import asyncio
import logging
import yaml

import click
import click_log

from robotica.executor import Executor
from robotica.inputs.http import HttpInput
from robotica.inputs.mqtt import MqttInput
from robotica.outputs.audio import AudioOutput
from robotica.outputs.lifx import LifxOutput
from robotica.outputs.mqtt import MqttOutput
from robotica.schedule import Schedule

logger = logging.getLogger('robotica')
click_log.basic_config(logger)


@click.command()
@click.option('--config', default="config/config.yaml", help='Path to config.')
@click.option('--schedule', default="config/schedule.yaml", help='Path to schedule config.')
@click_log.simple_verbosity_option(logger)
def main(config: str, schedule: str) -> None:
    """Console script for robotica."""
    with open(config, "r") as file:
        config_dict = yaml.safe_load(file)
        output_dict = config_dict['outputs']
        input_dict = config_dict['inputs']

    loop = asyncio.get_event_loop()

    lifx_output = LifxOutput(loop, output_dict['lifx'])
    lifx_output.start()

    audio_output = AudioOutput(loop, output_dict['audio'])
    audio_output.start()

    mqtt_output = MqttOutput(loop, output_dict['mqtt'])
    mqtt_output.start()

    executor_obj = Executor(loop, config_dict['executor'])
    executor_obj.add_output(audio_output)
    executor_obj.add_output(lifx_output)
    executor_obj.add_output(mqtt_output)

    schedule_obj = Schedule(schedule, executor_obj)
    schedule_obj.start()

    http_input = HttpInput(loop, input_dict['http'], executor_obj, schedule_obj)
    http_input.start()

    mqtt_input = MqttInput(loop, input_dict['mqtt'], executor_obj, schedule_obj)
    mqtt_input.start()

    try:
        loop.run_forever()
    finally:
        mqtt_input.stop()
        http_input.stop()
        schedule_obj.stop()
        mqtt_output.stop()
        audio_output.stop()
        lifx_output.stop()
        pending = asyncio.Task.all_tasks()
        for p in pending:
            p.cancel()
            try:
                loop.run_until_complete(p)
            except asyncio.CancelledError:
                pass
        loop.close()



import asyncio
import json
import logging
import platform

import yaml

from typing import Any, Optional, Dict

from hbmqtt.client import MQTTClient, ClientException, QOS_0

from robotica.executor import Executor
from robotica.inputs import Input
from robotica.schedule import Schedule

logger = logging.getLogger(__name__)

JsonType = Any


TOPICS = [
    ('/execute/', QOS_0),
    ('/schedule/', QOS_0),
]


class MqttInput(Input):
    def __init__(
            self, loop: asyncio.AbstractEventLoop,
            config: Dict,
            executor: Executor,
            schedule: Schedule) -> None:
        self._loop = loop
        self._config = config
        self._disabled = self._config['disabled']
        self._broker_url = self._config['broker_url']
        self._executor = executor
        self._schedule = schedule
        self._task = None  # type: Optional[asyncio.Task]
        self._client = MQTTClient()

    def start(self) -> None:
        if not self._disabled:
            self._task = self._loop.create_task(self._mqtt())

    def stop(self) -> None:
        if not self._disabled and self._task is not None:
            self._task.cancel()
            try:
                self._loop.run_until_complete(self._task)
            except asyncio.CancelledError:
                pass

    async def _execute(self, data: JsonType) -> None:

        reply_topic = data.get('reply_topic', None)
        server = platform.node()

        async def reply(data: JsonType) -> None:
            client = self._client
            raw_data = json.dumps(data).encode('UTF8')
            if reply_topic is not None:
                await client.publish(reply_topic, raw_data, qos=QOS_0)

        try:
            locations = data['locations']
            actions = data['actions']
        except KeyError:
            logger.error("Required value missing.")
            await reply({'status': 'error', 'server': server, })
            return

        try:
            await reply({'status': 'processing', 'server': server, })
            await self._executor.do_actions(locations, actions)
            await reply({'status': 'success', 'server': server, })
        except Exception as e:
            logger.exception("Error in _execute")
            await reply({'status': 'error', 'server': server, })

        return

    async def _process_schedule(self, data: JsonType) -> None:
         await self._schedule.set_schedule(data)
         self._schedule.save_schedule()

    async def _process(self, topic: str, data: JsonType) -> None:
        if topic.startswith("/execute/"):
            await self._execute(data)
        if topic.startswith("/schedule/"):
            await self._process_schedule(data)

    async def _mqtt(self) -> None:
        client = self._client
        await client.connect(self._broker_url)
        await client.subscribe(TOPICS)

        while True:
            try:
                message = await client.deliver_message()
                packet = message.publish_packet
                topic = packet.variable_header.topic_name
                raw_data = bytes(packet.payload.data).decode('UTF8')

                try:
                    data = json.loads(raw_data)
                    await self._process(topic, data)
                except json.JSONDecodeError as e:
                    logger.error("JSON Error %s" % e)

            except asyncio.CancelledError:
                await client.unsubscribe([t[0] for t in TOPICS])
                await client.disconnect()
                raise
            except ClientException as e:
                logger.error("Client exception: %s" % e)
            except Exception as e:
                logger.exception("Unknown exception.")



""" Give verbal message. """
import asyncio
import json

from hbmqtt.client import MQTTClient, ClientException, QOS_0
import logging
from typing import Set, Dict
import yaml

from robotica.outputs import Output
from robotica.types import JsonType, Action

logger = logging.getLogger(__name__)


class MqttOutput(Output):

    def __init__(self, loop: asyncio.AbstractEventLoop, config: Dict) -> None:
        self._loop = loop
        self._config = config
        self._disabled = self._config['disabled']
        self._broker_url = self._config['broker_url']
        self._locations = self._config.get('locations', {}) or {}
        self._client = MQTTClient()

    def start(self) -> None:
        if not self._disabled:
            self._loop.run_until_complete(self._client.connect(self._broker_url))

    def stop(self) -> None:
        pass

    def is_action_required_for_locations(self, locations: Set[str], action: Action) -> bool:
        if self._disabled:
            return False

        good_locations = [
            location
            for location in locations
            if location in self._locations
        ]

        if len(good_locations) == 0:
            return False

        if 'message' in action:
            return True

        if 'music' in action:
            return True

        return False

    async def execute(self, locations: Set[str], action: Action) -> None:
        if 'message' in action:
            message = action['message']

            await self.say(
                locations=locations,
                text=message['text'])

        if 'music' in action:
            music = action['music']

            await self.music_play(
                locations=locations,
                play_list=music['play_list'])

    async def _execute(self, topic: str, data: JsonType) -> None:
        logger.debug("About to publish %r to %s" % (data, topic))
        raw_data = json.dumps(data).encode('UTF8')
        try:
            await self._client.publish(
                topic,
                raw_data,
                qos=QOS_0,
            )
        except ClientException:
            logger.exception("The client operation failed.")

    async def _say(self, location: str, text: str) -> None:
        if location not in self._locations:
            return
        logger.debug("%s: About to say '%s' (MQTT).", location, text)
        await self._execute(
            '/say/%s/' % location,
            text,
        )

    async def say(self, locations: Set[str], text: str) -> None:
        if self._disabled:
            return
        coros = [
            self._say(location, text)
            for location in locations
        ]
        await asyncio.gather(*coros, loop=self._loop)

    async def _play(self, location: str, sound: str) -> None:
        if location not in self._locations:
            return
        logger.debug("%s: About to play sound '%s' (MQTT).", location, sound)
        await self._execute(
            '/play/%s/' % location,
            sound,
        )

    async def play(self, locations: Set[str], sound: str) -> None:
        if self._disabled:
            return
        coros = [
            self._play(location, sound)
            for location in locations
        ]
        await asyncio.gather(*coros, loop=self._loop)

    async def _music_play(self, location: str, play_list: str) -> None:
        if location not in self._locations:
            return
        logger.debug("%s: About to play music '%s' (MQTT).", location, play_list)
        await self._execute(
            '/play_music/%s/' % location,
            play_list,
        )

    async def music_play(self, locations: Set[str], play_list: str) -> None:
        if self._disabled:
            return
        coros = [
            self._music_play(location, play_list)
            for location in locations
        ]
        await asyncio.gather(*coros, loop=self._loop)

    async def _music_stop(self, location: str) -> None:
        if location not in self._locations:
            return
        logger.debug("%s: About to stop music (MQTT).", location)
        await self._execute(
            '/stop_music/%s/' % location,
            None,
        )

    async def music_stop(self, locations: Set[str]) -> None:
        if self._disabled:
            return
        coros = [
            self._music_stop(location)
            for location in locations
        ]
        await asyncio.gather(*coros, loop=self._loop)
